@@ -1,14 +1,13 @@
 /* io.c
  *
- * Copyright (c) 1996-2000 Mike Gleason, NCEMRSoft.
+ * Copyright (c) 1996-2001 Mike Gleason, NCEMRSoft.
  * All rights reserved.
  *
  */
 
 #include "syshdrs.h"
 
-static int gGotBrokenData;
-static int gUnused;
+static int gGotBrokenData = 0;
 
 #if defined(WIN32) || defined(_WINDOWS)
 #	define ASCII_TRANSLATION 0
@@ -29,7 +28,7 @@ static sigjmp_buf gBrokenDataJmp;
 #else
 static jmp_buf gBrokenDataJmp;
 #endif	/* HAVE_SIGSETJMP */
-static int gCanBrokenDataJmp;
+static int gCanBrokenDataJmp = 0;
 
 #endif	/* NO_SIGNALS */
 
@@ -43,8 +42,8 @@ static int gCanBrokenDataJmp;
 #	endif
 #endif
 
-static void WaitForRemoteInput(const FTPCIPtr cip);
-static void WaitForRemoteOutput(const FTPCIPtr cip);
+static int WaitForRemoteInput(const FTPCIPtr cip);
+static int WaitForRemoteOutput(const FTPCIPtr cip);
 
 
 #ifndef NO_SIGNALS
@@ -82,6 +81,7 @@ FTPInitIOTimer(const FTPCIPtr cip)
 	cip->secLeft = -1.0;
 	cip->nextProgressUpdate = 0;
 	cip->stalled = 0;
+	cip->dataTimedOut = 0;
 	cip->useProgressMeter = 1;
 	(void) gettimeofday(&cip->t0, NULL);
 }	/* FTPInitIOTimer */
@@ -174,8 +174,8 @@ int
 FTPList(const FTPCIPtr cip, const int outfd, const int longMode, const char *const lsflag)
 {
 	const char *cmd;
-	char line[128];
-	char secondaryBuf[512];
+	char line[512];
+	char secondaryBuf[768];
 #ifndef NO_SIGNALS
 	char *secBufPtr, *secBufLimit;
 	int nread;
@@ -201,7 +201,7 @@ FTPList(const FTPCIPtr cip, const int outfd, const int longMode, const char *con
 #ifdef NO_SIGNALS
 
 	if (result == 0) {
-		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout) < 0) {
+		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout, 1) < 0) {
 			/* Not really fdopen, but close in what we're trying to do. */
 			result = kErrFdopenR;
 			cip->errNo = kErrFdopenR;
@@ -407,6 +407,10 @@ FTPListToMemory2(const FTPCIPtr cip, const char *const pattern, const LineListPt
 			dcp = lsflags1;
 			lim = lsflags1 + sizeof(lsflags1) - 2;
 			for (; *scp != '\0'; scp++) {
+				if (isspace(*scp))
+					continue;
+				if (*scp == '-')
+					continue;
 				if (*scp == 'l') {
 					/* do not add the 'l' */
 					command = "LIST";
@@ -440,7 +444,7 @@ FTPListToMemory2(const FTPCIPtr cip, const char *const pattern, const LineListPt
 #ifdef NO_SIGNALS
 
 	if (result == 0) {
-		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout) < 0) {
+		if (InitSReadlineInfo(&lsSrl, cip->dataSocket, secondaryBuf, sizeof(secondaryBuf), (int) cip->xferTimeout, 1) < 0) {
 			/* Not really fdopen, but close in what we're trying to do. */
 			result = kErrFdopenR;
 			cip->errNo = kErrFdopenR;
@@ -561,10 +565,79 @@ FTPListToMemory2(const FTPCIPtr cip, const char *const pattern, const LineListPt
 
 
 
-/* The purpose of this is to provide updates for the progress meters
- * during lags.
- */
 static void
+AutomaticallyUseASCIIModeDependingOnExtension(const FTPCIPtr cip, const char *const pathName, int *const xtype)
+{
+	if ((*xtype == kTypeBinary) && (cip->asciiFilenameExtensions != NULL)) {
+		if (FilenameExtensionIndicatesASCII(pathName, cip->asciiFilenameExtensions)) {
+			/* Matched -- send this file in ASCII mode
+			 * instead of binary since it's extension
+			 * appears to be that of a text file.
+			 */
+			*xtype = kTypeAscii;
+		}
+	}
+}	/* AutomaticallyUseASCIIModeDependingOnExtension */
+
+
+
+static void
+FTPCheckForRestartModeAvailability(const FTPCIPtr cip)
+{
+	if (cip->hasREST == kCommandAvailabilityUnknown) {
+		(void) FTPSetTransferType(cip, kTypeBinary);
+		if (SetStartOffset(cip, (longest_int) 1) == kNoErr) {
+			/* Now revert -- we still may not end up
+			 * doing it.
+			 */
+			SetStartOffset(cip, (longest_int) -1);
+		}
+	}
+}	/* FTPCheckForRestartModeAvailability */
+
+
+
+
+static void
+FTPSetUploadSocketBufferSize(const FTPCIPtr cip)
+{
+	/* If dataSocketSBufSize is non-zero, it means you
+	 * want to explicitly try to set the size of the
+	 * socket's I/O buffer.
+	 *
+	 * If it is zero, it means you want to just use the
+	 * TCP stack's default value, which is typically
+	 * between 8 and 64 kB.
+	 *
+	 * If you try to set the buffer larger than 64 kB,
+	 * the TCP stack should try to use RFC 1323 to
+	 * negotiate "TCP Large Windows" which may yield
+	 * significant performance gains.
+	 */
+	if ((cip->numUploads == 0) && (cip->dataSocketSBufSize > 0)) {
+		if (cip->hasSTORBUFSIZE == kCommandAvailable)
+			(void) FTPCmd(cip, "SITE STORBUFSIZE %lu", (unsigned long) cip->dataSocketSBufSize);
+		else if (cip->hasSBUFSIZ == kCommandAvailable)
+			(void) FTPCmd(cip, "SITE SBUFSIZ %lu", (unsigned long) cip->dataSocketSBufSize);
+		else if (cip->hasSBUFSZ == kCommandAvailable)
+			(void) FTPCmd(cip, "SITE SBUFSZ %lu", (unsigned long) cip->dataSocketSBufSize);
+		/* At least one server implemenation has RBUFSZ but not
+		 * SBUFSZ and instead uses RBUFSZ for both.
+		 */
+		else if ((cip->hasSBUFSZ != kCommandAvailable) && (cip->hasRBUFSZ == kCommandAvailable))
+			(void) FTPCmd(cip, "SITE RBUFSZ %lu", (unsigned long) cip->dataSocketSBufSize);
+		else if (cip->hasBUFSIZE == kCommandAvailable)
+			(void) FTPCmd(cip, "SITE BUFSIZE %lu", (unsigned long) cip->dataSocketSBufSize);
+	}
+}	/* FTPSetUploadSocketBufferSize */
+
+
+
+
+/* The purpose of this is to provide updates for the progress meters
+ * during lags.  Return zero if the operation timed-out.
+ */
+static int
 WaitForRemoteOutput(const FTPCIPtr cip)
 {
 	fd_set ss, ss2;
@@ -576,19 +649,22 @@ WaitForRemoteOutput(const FTPCIPtr cip)
 	int ocancelXfer;
 
 	xferTimeout = cip->xferTimeout;
-	if (xferTimeout == 1)
-		return;
+	if (xferTimeout < 1)
+		return (1);
 
 	fd = cip->dataSocket;
 	if (fd < 0)
-		return;
+		return (1);
 
 	ocancelXfer = cip->cancelXfer;
 	wsecs = 0;
+	cip->stalled = 0;
 
 	while ((xferTimeout <= 0) || (wsecs < xferTimeout)) {
-		if ((cip->cancelXfer != 0) && (ocancelXfer == 0))
-			return;
+		if ((cip->cancelXfer != 0) && (ocancelXfer == 0)) {
+			/* leave cip->stalled -- could have been stalled and then canceled. */
+			return (1);
+		}
 		FD_ZERO(&ss);
 		FD_SET(fd, &ss);
 		ss2 = ss;
@@ -598,16 +674,16 @@ WaitForRemoteOutput(const FTPCIPtr cip)
 		if (result == 1) {
 			/* ready */
 			cip->stalled = 0;
-			return;
+			return (1);
 		} else if (result < 0) {
-			cip->stalled = 0;
 			if (errno != EINTR) {
 				perror("select");
-				return;
+				cip->stalled = 0;
+				return (1);
 			}
 		} else {
-			cip->stalled++;
 			wsecs++;
+			cip->stalled = wsecs;
 		}
 		FTPUpdateIOTimer(cip);
 	}
@@ -618,6 +694,9 @@ WaitForRemoteOutput(const FTPCIPtr cip)
 	 */
 	(void) kill(getpid(), SIGALRM);
 #endif	/* NO_SIGNALS */
+
+	cip->dataTimedOut = 1;
+	return (0);	/* timed-out */
 }	/* WaitForRemoteOutput */
 
 
@@ -628,7 +707,7 @@ FTPPutOneF(
 	const FTPCIPtr cip,
 	const char *const file,
 	const char *volatile dstfile,
-	const int xtype,
+	int xtype,
 	const int fdtouse,
 	const int appendflag,
 	const char *volatile tmppfx,
@@ -653,7 +732,7 @@ FTPPutOneF(
 #endif
 	int fstatrc, statrc;
 	longest_int startPoint = 0;
-	struct stat st;
+	struct Stat st;
 	time_t mdtm;
 #if !defined(NO_SIGNALS)
 	int sj;
@@ -672,7 +751,7 @@ FTPPutOneF(
 
 	cip->usingTAR = 0;
 	if (fdtouse < 0) {
-		fd = open(file, O_RDONLY|O_BINARY, 0);
+		fd = Open(file, O_RDONLY|O_BINARY, 0);
 		if (fd < 0) {
 			Error(cip, kDoPerror, "Cannot open local file %s for reading.\n", file);
 			cip->errNo = kErrOpenFailed;
@@ -682,7 +761,7 @@ FTPPutOneF(
 		fd = fdtouse;
 	}
 
-	fstatrc = fstat(fd, &st);
+	fstatrc = Fstat(fd, &st);
 	if ((fstatrc == 0) && (S_ISDIR(st.st_mode))) {
 		if (fdtouse < 0) {
 			(void) close(fd);
@@ -695,17 +774,12 @@ FTPPutOneF(
 	/* For Put, we can't recover very well if it turns out restart
 	 * didn't work, so check beforehand.
 	 */
-	if (cip->hasREST == kCommandAvailabilityUnknown) {
-		(void) FTPSetTransferType(cip, kTypeBinary);
-		if (SetStartOffset(cip, (longest_int) 1) == kNoErr) {
-			/* Now revert -- we still may not end up
-			 * doing it.
-			 */
-			SetStartOffset(cip, (longest_int) -1);
-		}
+	if ((resumeflag == kResumeYes) || (resumeProc != NoConfirmResumeUploadProc)) {
+		FTPCheckForRestartModeAvailability(cip); 
 	}
 
 	if (fdtouse < 0) {
+		AutomaticallyUseASCIIModeDependingOnExtension(cip, dstfile, &xtype);
 		(void) FTPFileSizeAndModificationTime(cip, dstfile, &startPoint, xtype, &mdtm);
 
 		if (appendflag == kAppendYes) {
@@ -723,10 +797,20 @@ FTPPutOneF(
 		}
 
 		statrc = -1;
-		if ((mdtm != kModTimeUnknown) || (startPoint != kSizeUnknown))
+		if ((mdtm != kModTimeUnknown) || (startPoint != kSizeUnknown)) {
+			/* Then we know the file exists.  We will
+			 * ask the user what to do, if possible, below.
+			 */
 			statrc = 0;
-		else if (resumeProc != NoConfirmResumeUploadProc)
+		} else if ((resumeProc != NoConfirmResumeUploadProc) && (cip->hasMDTM != kCommandAvailable) && (cip->hasSIZE != kCommandAvailable)) {
+			/* We already checked if the file had a filesize
+			 * or timestamp above, but if the server indicated
+			 * it did not support querying those directly,
+			 * we now need to try to determine if the file
+			 * exists in a few other ways.
+			 */
 			statrc = FTPFileExists2(cip, dstfile, 0, 0, 0, 1, 1);
+		}
 
 		if (
 			(resumeProc != NoConfirmResumeUploadProc) &&
@@ -791,26 +875,19 @@ FTPPutOneF(
 					}
 				}
 				return (kNoErr);
-			} else if (lseek(fd, (off_t) startPoint, SEEK_SET) != (off_t) -1) {
+			} else if (Lseek(fd, (off_t) startPoint, SEEK_SET) != (off_t) -1) {
 				cip->startPoint = startPoint;
 			}
 		} else if (zaction == kConfirmResumeProcSaidAppend) {
 			/* append: leave startPoint at zero, we will append everything. */
-			cip->startPoint = startPoint;
+			cip->startPoint = startPoint = 0;
 		} else /* if (zaction == kConfirmResumeProcSaidOverwrite) */ {
 			/* overwrite: leave startPoint at zero */
 			cip->startPoint = startPoint = 0;
 		}
 	}
 
-	if ((cip->numUploads == 0) && (cip->dataSocketSBufSize > 0)) {
-		if (cip->hasSTORBUFSIZE == kCommandAvailable)
-			(void) FTPCmd(cip, "SITE STORBUFSIZE %lu", (unsigned long) cip->dataSocketSBufSize);
-		if (cip->hasSBUFSIZ == kCommandAvailable)
-			(void) FTPCmd(cip, "SITE SBUFSIZ %lu", (unsigned long) cip->dataSocketSBufSize);
-		if (cip->hasSBUFSZ == kCommandAvailable)
-			(void) FTPCmd(cip, "SITE SBUFSZ %lu", (unsigned long) cip->dataSocketSBufSize);
-	}
+	FTPSetUploadSocketBufferSize(cip);
 
 #ifdef NO_SIGNALS
 	vzaction = zaction;
@@ -902,7 +979,7 @@ FTPPutOneF(
 		 *
 		 * So now we have to undo our seek.
 		 */
-		if (lseek(fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
+		if (Lseek(fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
 			cip->errNo = kErrLseekFailed;
 			if (fdtouse < 0) {
 				(void) close(fd);
@@ -976,7 +1053,11 @@ FTPPutOneF(
 				(void) alarm(cip->xferTimeout);
 #endif	/* NO_SIGNALS */
 			do {
-				WaitForRemoteOutput(cip);		/* may set cancelXfer */
+				if (! WaitForRemoteOutput(cip)) {	/* could set cancelXfer */
+					cip->errNo = result = kErrDataTimedOut;
+					Error(cip, kDontPerror, "Remote write timed out.\n");
+					goto brk;
+				}
 				if (cip->cancelXfer > 0) {
 					FTPAbortDataTransfer(cip);
 					result = cip->errNo = kErrDataTransferAborted;
@@ -1054,7 +1135,11 @@ FTPPutOneF(
 				(void) alarm(cip->xferTimeout);
 #endif	/* NO_SIGNALS */
 			do {
-				WaitForRemoteOutput(cip);		/* may set cancelXfer */
+				if (! WaitForRemoteOutput(cip)) {	/* could set cancelXfer */
+					cip->errNo = result = kErrDataTimedOut;
+					Error(cip, kDontPerror, "Remote write timed out.\n");
+					goto brk;
+				}
 				if (cip->cancelXfer > 0) {
 					FTPAbortDataTransfer(cip);
 					result = cip->errNo = kErrDataTransferAborted;
@@ -1078,6 +1163,7 @@ FTPPutOneF(
 						Error(cip, kDoPerror, "Remote write failed.\n");
 					}
 					(void) shutdown(cip->dataSocket, 2);
+					cip->dataSocket = -1;
 					goto brk;
 				}
 #else	/* NO_SIGNALS */
@@ -1094,6 +1180,7 @@ FTPPutOneF(
 						Error(cip, kDoPerror, "Remote write failed.\n");
 					}
 					(void) shutdown(cip->dataSocket, 2);
+					cip->dataSocket = -1;
 					goto brk;
 				}
 #endif	/* NO_SIGNALS */
@@ -1106,10 +1193,9 @@ FTPPutOneF(
 brk:
 
 	if (fdtouse < 0) {
-		(void) fstat(fd, &st);
+		(void) Fstat(fd, &st);
 	}
 
-	WaitForRemoteOutput(cip);	/* Close could block. */
 	if (fdtouse < 0) {
 		if (shutdown(fd, 1) == 0) {
 			/* This looks very bizarre, since
@@ -1126,7 +1212,7 @@ brk:
 			 * the whole socket, which tells
 			 * us that the file was completed.
 			 */
-			WaitForRemoteInput(cip);	/* Close could block. */
+			(void) WaitForRemoteInput(cip);	/* Close could block. */
 		}
 	}
 
@@ -1147,6 +1233,7 @@ brk:
 		 * we need to dispose of it.
 		 */
 		(void) close(fd);
+		fd = -1;
 	}
 
 	if (result == kNoErr) {
@@ -1178,7 +1265,7 @@ brk:
 				Error(cip, kDontPerror, "Could not preserve times for %s: %s.\n", odstfile, FTPStrError(cip->errNo));
 		}
 
-		if (deleteflag == kDeleteYes) {
+		if ((result == kNoErr) && (deleteflag == kDeleteYes)) {
 			if (unlink(file) < 0) {
 				result = cip->errNo = kErrLocalDeleteFailed;
 			}
@@ -1207,11 +1294,11 @@ FTPPutOneFile3(
 	const int resumeflag,
 	const int deleteflag,
 	const ConfirmResumeUploadProc resumeProc,
-	int reserved)
+	int UNUSED(reserved))
 {
 	int result;
 
-	gUnused = reserved;
+	LIBNCFTP_USE_VAR(reserved);
 	if (cip == NULL)
 		return (kErrBadParameter);
 	if (strcmp(cip->magic, kLibraryMagic))
@@ -1244,7 +1331,7 @@ FTPPutFiles3(
 	const int resumeflag,
 	const int deleteflag,
 	const ConfirmResumeUploadProc resumeProc,
-	int reserved)
+	int UNUSED(reserved))
 {
 	LineList globList;
 	FileInfoList files;
@@ -1254,7 +1341,7 @@ FTPPutFiles3(
 	const char *dstdir;
 	char dstdir2[512];
 
-	gUnused = reserved;
+	LIBNCFTP_USE_VAR(reserved);
 	if (cip == NULL)
 		return (kErrBadParameter);
 	if (strcmp(cip->magic, kLibraryMagic))
@@ -1352,11 +1439,225 @@ FTPPutFiles3(
 
 
 
+int
+FTPPutFileFromMemory(
+	const FTPCIPtr cip,
+	const char *volatile dstfile,
+	const char *volatile src,
+	const size_t srcLen,
+	const int appendflag)
+{
+	char *buf, *cp;
+	const char *cmd;
+	int tmpResult, result;
+	int nread, nwrote;
+	size_t bufSize;
+	const char *srcLim;
+	const char *volatile srcp;
+#if !defined(NO_SIGNALS)
+	int sj;
+	volatile FTPSigProc osigpipe;
+	volatile FTPCIPtr vcip;
+#endif	/* NO_SIGNALS */
+
+	if (cip->buf == NULL) {
+		Error(cip, kDoPerror, "Transfer buffer not allocated.\n");
+		cip->errNo = kErrNoBuf;
+		return (cip->errNo);
+	}
+
+	cip->usingTAR = 0;
+
+	/* For Put, we can't recover very well if it turns out restart
+	 * didn't work, so check beforehand.
+	 */
+	FTPCheckForRestartModeAvailability(cip); 
+
+	FTPSetUploadSocketBufferSize(cip);
+
+#ifdef NO_SIGNALS
+#else	/* NO_SIGNALS */
+	vcip = cip;
+	osigpipe = (volatile FTPSigProc) signal(SIGPIPE, BrokenData);
+
+	gGotBrokenData = 0;
+	gCanBrokenDataJmp = 0;
+
+#ifdef HAVE_SIGSETJMP
+	sj = sigsetjmp(gBrokenDataJmp, 1);
+#else
+	sj = setjmp(gBrokenDataJmp);
+#endif	/* HAVE_SIGSETJMP */
+
+	if (sj != 0) {
+		(void) signal(SIGPIPE, (FTPSigProc) osigpipe);
+		FTPShutdownHost(vcip);
+		vcip->errNo = kErrRemoteHostClosedConnection;
+		return(vcip->errNo);
+	}
+	gCanBrokenDataJmp = 1;
+#endif	/* NO_SIGNALS */
+
+	cmd = appendflag == kAppendYes ? "APPE" : "STOR";
+	tmpResult = FTPStartDataCmd(
+		cip,
+		kNetWriting,
+		kTypeBinary,
+		(longest_int) 0,
+		"%s %s",
+		cmd,
+		dstfile
+	);
+
+	if (tmpResult < 0) {
+		cip->errNo = tmpResult;
+#if !defined(NO_SIGNALS)
+		(void) signal(SIGPIPE, (FTPSigProc) osigpipe);
+#endif	/* NO_SIGNALS */
+		return (cip->errNo);
+	}
+
+	result = kNoErr;
+	buf = cip->buf;
+	bufSize = cip->bufSize;
+
+	FTPInitIOTimer(cip);
+	cip->expectedSize = srcLen;
+	cip->lname = NULL;	/* could be NULL */
+	cip->rname = dstfile;
+	srcp = src;
+	srcLim = src + srcLen;
+	FTPStartIOTimer(cip);
+	{
+		/* binary */
+		for (;;) {
+#if !defined(NO_SIGNALS)
+			gCanBrokenDataJmp = 0;
+#endif	/* NO_SIGNALS */
+			nread = bufSize;
+			if ((size_t) (srcLim - srcp) < bufSize) {
+				nread = (int) (srcLim - srcp);
+				if (nread == 0) {
+					result = kNoErr;
+					break;
+				}
+			}
+			cip->bytesTransferred += (longest_int) nread;
+			cp = srcp;
+			srcp += nread;
+
+#if !defined(NO_SIGNALS)
+			gCanBrokenDataJmp = 1;
+			if (cip->xferTimeout > 0)
+				(void) alarm(cip->xferTimeout);
+#endif	/* NO_SIGNALS */
+			do {
+				if (! WaitForRemoteOutput(cip)) {	/* could set cancelXfer */
+					cip->errNo = result = kErrDataTimedOut;
+					Error(cip, kDontPerror, "Remote write timed out.\n");
+					goto brk;
+				}
+				if (cip->cancelXfer > 0) {
+					FTPAbortDataTransfer(cip);
+					result = cip->errNo = kErrDataTransferAborted;
+					goto brk;
+				}
+
+#ifdef NO_SIGNALS
+				nwrote = SWrite(cip->dataSocket, cp, (size_t) nread, (int) cip->xferTimeout, kNoFirstSelect);
+				if (nwrote < 0) {
+					if (nwrote == kTimeoutErr) {
+						cip->errNo = result = kErrDataTimedOut;
+						Error(cip, kDontPerror, "Remote write timed out.\n");
+					} else if ((gGotBrokenData != 0) || (errno == EPIPE)) {
+						cip->errNo = result = kErrSocketWriteFailed;
+						errno = EPIPE;
+						Error(cip, kDoPerror, "Lost data connection to remote host.\n");
+					} else if (errno == EINTR) {
+						continue;
+					} else {
+						cip->errNo = result = kErrSocketWriteFailed;
+						Error(cip, kDoPerror, "Remote write failed.\n");
+					}
+					(void) shutdown(cip->dataSocket, 2);
+					cip->dataSocket = -1;
+					goto brk;
+				}
+#else	/* NO_SIGNALS */
+				nwrote = write(cip->dataSocket, cp, nread);
+				if (nwrote < 0) {
+					if ((gGotBrokenData != 0) || (errno == EPIPE)) {
+						cip->errNo = result = kErrSocketWriteFailed;
+						errno = EPIPE;
+						Error(cip, kDoPerror, "Lost data connection to remote host.\n");
+					} else if (errno == EINTR) {
+						continue;
+					} else {
+						cip->errNo = result = kErrSocketWriteFailed;
+						Error(cip, kDoPerror, "Remote write failed.\n");
+					}
+					(void) shutdown(cip->dataSocket, 2);
+					cip->dataSocket = -1;
+					goto brk;
+				}
+#endif	/* NO_SIGNALS */
+				cp += nwrote;
+				nread -= nwrote;
+			} while (nread > 0);
+			FTPUpdateIOTimer(cip);
+		}
+	}
+brk:
+	if (shutdown(cip->dataSocket, 1) == 0) {
+		/* This looks very bizarre, since
+		 * we will be checking the socket
+		 * for readability here!
+		 *
+		 * The reason for this is that we
+		 * want to be able to timeout a
+		 * small put.  So, we close the
+		 * write end of the socket first,
+		 * which tells the server we're
+		 * done writing.  We then wait
+		 * for the server to close down
+		 * the whole socket, which tells
+		 * us that the file was completed.
+		 */
+		(void) WaitForRemoteInput(cip);	/* Close could block. */
+	}
+
+#if !defined(NO_SIGNALS)
+	gCanBrokenDataJmp = 0;
+	if (cip->xferTimeout > 0)
+		(void) alarm(0);
+#endif	/* NO_SIGNALS */
+	tmpResult = FTPEndDataCmd(cip, 1);
+	if ((tmpResult < 0) && (result == kNoErr)) {
+		cip->errNo = result = kErrSTORFailed;
+	}
+	FTPStopIOTimer(cip);
+
+	if (result == kNoErr) {
+		/* The store succeeded;  If we were
+		 * uploading to a temporary file,
+		 * move the new file to the new name.
+		 */
+		cip->numUploads++;
+	}
+
+#if !defined(NO_SIGNALS)
+	(void) signal(SIGPIPE, (FTPSigProc) osigpipe);
+#endif	/* NO_SIGNALS */
+	return (result);
+}	/* FTPPutFileFromMemory */
+
+
+
 
 /* The purpose of this is to provide updates for the progress meters
- * during lags.
+ * during lags.  Return zero if the operation timed-out.
  */
-static void
+static int
 WaitForRemoteInput(const FTPCIPtr cip)
 {
 	fd_set ss, ss2;
@@ -1368,19 +1669,22 @@ WaitForRemoteInput(const FTPCIPtr cip)
 	int ocancelXfer;
 
 	xferTimeout = cip->xferTimeout;
-	if (xferTimeout == 1)
-		return;
+	if (xferTimeout < 1)
+		return (1);
 
 	fd = cip->dataSocket;
 	if (fd < 0)
-		return;
+		return (1);
 
 	ocancelXfer = cip->cancelXfer;
 	wsecs = 0;
+	cip->stalled = 0;
 
 	while ((xferTimeout <= 0) || (wsecs < xferTimeout)) {
-		if ((cip->cancelXfer != 0) && (ocancelXfer == 0))
-			return;
+		if ((cip->cancelXfer != 0) && (ocancelXfer == 0)) {
+			/* leave cip->stalled -- could have been stalled and then canceled. */
+			return (1);
+		}
 		FD_ZERO(&ss);
 		FD_SET(fd, &ss);
 		ss2 = ss;
@@ -1389,14 +1693,17 @@ WaitForRemoteInput(const FTPCIPtr cip)
 		result = select(fd + 1, SELECT_TYPE_ARG234 &ss, NULL, SELECT_TYPE_ARG234 &ss2, &tv);
 		if (result == 1) {
 			/* ready */
-			return;
+			cip->stalled = 0;
+			return (1);
 		} else if (result < 0) {
 			if (result != EINTR) {
 				perror("select");
-				return;
+				cip->stalled = 0;
+				return (1);
 			}
 		} else {
 			wsecs++;
+			cip->stalled = wsecs;
 		}
 		FTPUpdateIOTimer(cip);
 	}
@@ -1407,6 +1714,9 @@ WaitForRemoteInput(const FTPCIPtr cip)
 	 */
 	(void) kill(getpid(), SIGALRM);
 #endif	/* NO_SIGNALS */
+
+	cip->dataTimedOut = 1;
+	return (0);	/* timed-out */
 }	/* WaitForRemoteInput */
 
 
@@ -1481,12 +1791,12 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 	volatile int result;
 	int nread, nwrote;
 	volatile int fd;
-	volatile FTPSigProc osigpipe;
-	volatile FTPCIPtr vcip;
 	volatile int vfd;
 	const char *volatile vfile;
 #ifndef NO_SIGNALS
 	int sj;
+	volatile FTPSigProc osigpipe;
+	volatile FTPCIPtr vcip;
 #endif
 	int pid, status;
 	char savedCwd[512];
@@ -1548,11 +1858,11 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 		return (result);
 	}
 
-	vcip = cip;
 	vfd = fd;
 	vfile = file;
 
 #ifndef NO_SIGNALS
+	vcip = cip;
 	osigpipe = (volatile FTPSigProc) signal(SIGPIPE, BrokenData);
 
 	gGotBrokenData = 0;
@@ -1598,7 +1908,9 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 			result = kErrRETRFailed;
 		cip->errNo = result;
 
+#ifndef NO_SIGNALS
 		(void) signal(SIGPIPE, SIG_IGN);
+#endif
 		(void) close(vfd);
 		for (;;) {
 #ifdef HAVE_WAITPID
@@ -1612,7 +1924,9 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 				break;		/* done */
 		}
 
+#ifndef NO_SIGNALS
 		(void) signal(SIGPIPE, (FTPSigProc) osigpipe);
+#endif
 		if (basecp != NULL)
 			(void) FTPChdir(cip, savedCwd);
 		return (result);
@@ -1629,7 +1943,11 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 
 	/* Binary */
 	for (;;) {
-		WaitForRemoteInput(cip);		/* could set cancelXfer */
+		if (! WaitForRemoteInput(cip)) {	/* could set cancelXfer */
+			cip->errNo = result = kErrDataTimedOut;
+			Error(cip, kDontPerror, "Remote read timed out.\n");
+			break;
+		}
 		if (cip->cancelXfer > 0) {
 			FTPAbortDataTransfer(cip);
 			result = cip->errNo = kErrDataTransferAborted;
@@ -1644,7 +1962,7 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 		nread = SRead(cip->dataSocket, buf, bufSize, (int) cip->xferTimeout, kFullBufferNotRequired|kNoFirstSelect);
 		if (nread == kTimeoutErr) {
 			cip->errNo = result = kErrDataTimedOut;
-			Error(cip, kDontPerror, "Remote write timed out.\n");
+			Error(cip, kDontPerror, "Remote read timed out.\n");
 			break;
 		} else if (nread < 0) {
 			if (errno == EINTR)
@@ -1713,7 +2031,9 @@ FTPGetOneTarF(const FTPCIPtr cip, const char *file, const char *const dstdir)
 		cip->errNo = kErrRETRFailed;
 	}
 	FTPStopIOTimer(cip);
+#if !defined(NO_SIGNALS)
 	(void) signal(SIGPIPE, (FTPSigProc) osigpipe);
+#endif
 
 	if ((result == 0) && (cip->bytesTransferred == 0)) {
 		result = kErrOpenFailed;
@@ -1735,7 +2055,7 @@ FTPGetOneF(
 	const FTPCIPtr cip,
 	const char *const file,
 	const char *dstfile,
-	const int xtype,
+	int xtype,
 	const int fdtouse,
 	longest_int expectedSize,
 	time_t mdtm,
@@ -1757,7 +2077,7 @@ FTPGetOneF(
 #endif
 	volatile longest_int startPoint = 0;
 	struct utimbuf ut;
-	struct stat st;
+	struct Stat st;
 #if !defined(NO_SIGNALS)
 	volatile FTPSigProc osigpipe;
 	volatile FTPCIPtr vcip;
@@ -1794,6 +2114,7 @@ FTPGetOneF(
 		 *
 		 */
 
+		AutomaticallyUseASCIIModeDependingOnExtension(cip, file, &xtype);
 		if (expectedSize == kSizeUnknown) {
 			(void) FTPFileSizeAndModificationTime(cip, file, &expectedSize, xtype, &mdtm);
 		} else {
@@ -1804,15 +2125,7 @@ FTPGetOneF(
 		 * didn't work, so check beforehand.
 		 */
 		if ((resumeflag == kResumeYes) || (resumeProc != NoConfirmResumeDownloadProc)) {
-			if (cip->hasREST == kCommandAvailabilityUnknown) {
-				(void) FTPSetTransferType(cip, kTypeBinary);
-				if (SetStartOffset(cip, (longest_int) 1) == kNoErr) {
-					/* Now revert -- we still may not end up
-					 * doing it.
-					 */
-					SetStartOffset(cip, (longest_int) -1);
-				}
-			}
+			FTPCheckForRestartModeAvailability(cip); 
 		}
 
 		if (appendflag == kAppendYes) {
@@ -1825,7 +2138,7 @@ FTPGetOneF(
 			zaction = kConfirmResumeProcSaidOverwrite;
 		}
 
-		statrc = stat(dstfile, &st);
+		statrc = Stat(dstfile, &st);
 		if (statrc == 0) {
 			if (resumeProc != NULL) {
 				zaction = (*resumeProc)(
@@ -1953,15 +2266,15 @@ FTPGetOneF(
 				cip->errNo = result = kErrSetStartPoint;
 				return (result);
 			}
-			fd = open(dstfile, O_WRONLY|O_APPEND|O_BINARY, 00666);
+			fd = Open(dstfile, O_WRONLY|O_APPEND|O_BINARY, 00666);
 		} else if (zaction == kConfirmResumeProcSaidAppend) {
 			/* leave startPoint at zero, we will append everything. */
 			startPoint = (longest_int) 0;
-			fd = open(dstfile, O_WRONLY|O_CREAT|O_APPEND|O_BINARY, 00666);
+			fd = Open(dstfile, O_WRONLY|O_CREAT|O_APPEND|O_BINARY, 00666);
 		} else /* if (zaction == kConfirmResumeProcSaidOverwrite) */ {
 			created = 1;
 			startPoint = (longest_int) 0;
-			fd = open(dstfile, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 00666);
+			fd = Open(dstfile, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 00666);
 		}
 
 		if (fd < 0) {
@@ -1971,7 +2284,7 @@ FTPGetOneF(
 			return (result);
 		}
 
-		if ((expectedSize == (longest_int) 0) && (startPoint <= (longest_int) 0)) {
+		if ((expectedSize == (longest_int) 0) && (startPoint <= (longest_int) 0) && (zaction != kConfirmResumeProcSaidOverwrite)) {
 			/* Don't go to all the trouble of downloading nothing. */
 #if defined(WIN32) || defined(_WINDOWS)
 			/* Note: Windows doesn't allow zero-size files. */
@@ -1993,12 +2306,27 @@ FTPGetOneF(
 	}
 
 	if ((cip->numDownloads == 0) && (cip->dataSocketRBufSize > 0)) {
+		/* If dataSocketSBufSize is non-zero, it means you
+		 * want to explicitly try to set the size of the
+		 * socket's I/O buffer.
+		 *
+		 * If it is zero, it means you want to just use the
+		 * TCP stack's default value, which is typically
+		 * between 8 and 64 kB.
+		 *
+		 * If you try to set the buffer larger than 64 kB,
+		 * the TCP stack should try to use RFC 1323 to
+		 * negotiate "TCP Large Windows" which may yield
+		 * significant performance gains.
+		 */
 		if (cip->hasRETRBUFSIZE == kCommandAvailable)
 			(void) FTPCmd(cip, "SITE RETRBUFSIZE %lu", (unsigned long) cip->dataSocketRBufSize);
-		if (cip->hasRBUFSIZ == kCommandAvailable)
+		else if (cip->hasRBUFSIZ == kCommandAvailable)
 			(void) FTPCmd(cip, "SITE RBUFSIZ %lu", (unsigned long) cip->dataSocketRBufSize);
-		if (cip->hasRBUFSZ == kCommandAvailable)
+		else if (cip->hasRBUFSZ == kCommandAvailable)
 			(void) FTPCmd(cip, "SITE RBUFSZ %lu", (unsigned long) cip->dataSocketRBufSize);
+		else if (cip->hasBUFSIZE == kCommandAvailable)
+			(void) FTPCmd(cip, "SITE BUFSIZE %lu", (unsigned long) cip->dataSocketSBufSize);
 	}
 
 #ifdef NO_SIGNALS
@@ -2053,7 +2381,7 @@ FTPGetOneF(
 		 *
 		 * So now we have to undo our seek.
 		 */
-		if (lseek(fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
+		if (Lseek(fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
 			cip->errNo = kErrLseekFailed;
 			if (fdtouse < 0) {
 				(void) close(fd);
@@ -2084,7 +2412,11 @@ FTPGetOneF(
 	if (xtype == kTypeAscii) {
 		/* Ascii */
 		for (;;) {
-			WaitForRemoteInput(cip);		/* could set cancelXfer */
+			if (! WaitForRemoteInput(cip)) {	/* could set cancelXfer */
+				cip->errNo = result = kErrDataTimedOut;
+				Error(cip, kDontPerror, "Remote read timed out.\n");
+				break;
+			}
 			if (cip->cancelXfer > 0) {
 				FTPAbortDataTransfer(cip);
 				result = cip->errNo = kErrDataTransferAborted;
@@ -2102,7 +2434,7 @@ FTPGetOneF(
 			nread = SRead(cip->dataSocket, buf, bufSize, (int) cip->xferTimeout, kFullBufferNotRequired|kNoFirstSelect);
 			if (nread == kTimeoutErr) {
 				cip->errNo = result = kErrDataTimedOut;
-				Error(cip, kDontPerror, "Remote write timed out.\n");
+				Error(cip, kDontPerror, "Remote read timed out.\n");
 				break;
 			} else if (nread < 0) {
 				if ((gGotBrokenData != 0) || (errno == EPIPE)) {
@@ -2206,7 +2538,11 @@ FTPGetOneF(
 	{
 		/* Binary */
 		for (;;) {
-			WaitForRemoteInput(cip);		/* could set cancelXfer */
+			if (! WaitForRemoteInput(cip)) {	/* could set cancelXfer */
+				cip->errNo = result = kErrDataTimedOut;
+				Error(cip, kDontPerror, "Remote read timed out.\n");
+				break;
+			}
 			if (cip->cancelXfer > 0) {
 				FTPAbortDataTransfer(cip);
 				result = cip->errNo = kErrDataTransferAborted;
@@ -2224,7 +2560,7 @@ FTPGetOneF(
 			nread = SRead(cip->dataSocket, buf, bufSize, (int) cip->xferTimeout, kFullBufferNotRequired|kNoFirstSelect);
 			if (nread == kTimeoutErr) {
 				cip->errNo = result = kErrDataTimedOut;
-				Error(cip, kDontPerror, "Remote write timed out.\n");
+				Error(cip, kDontPerror, "Remote read timed out.\n");
 				break;
 			} else if (nread < 0) {
 				if ((gGotBrokenData != 0) || (errno == EPIPE)) {
@@ -2309,7 +2645,9 @@ brk:
 		 * we need to close it.
 		 */
 		(void) close(fd);
+		fd = -1;
 	}
+
 	tmpResult = FTPEndDataCmd(cip, 1);
 	if ((tmpResult < 0) && (result == 0)) {
 		result = kErrRETRFailed;
@@ -2349,11 +2687,11 @@ FTPGetOneFile3(
 	const int appendflag,
 	const int deleteflag,
 	const ConfirmResumeDownloadProc resumeProc,
-	int reserved)
+	int UNUSED(reserved))
 {
 	int result;
 
-	gUnused = reserved;
+	LIBNCFTP_USE_VAR(reserved);
 	if (cip == NULL)
 		return (kErrBadParameter);
 	if (strcmp(cip->magic, kLibraryMagic))
@@ -2386,7 +2724,7 @@ FTPGetFiles3(
 	const int deleteflag,
 	const int tarflag,
 	const ConfirmResumeDownloadProc resumeProc,
-	int reserved)
+	int UNUSED(reserved))
 {
 	LineList globList;
 	LinePtr itemPtr;
@@ -2397,13 +2735,13 @@ FTPGetFiles3(
 	char *ldir;
 	char *cp;
 	const char *dstdir;
-	char dstdir2[512];
 	const char *pattern;
-	char pattern2[256];
+	char *pattern2, *dstdir2;
 	char c;
 	int recurse1;
+	int errRc;
 
-	gUnused = reserved;
+	LIBNCFTP_USE_VAR(reserved);
 	if (cip == NULL)
 		return (kErrBadParameter);
 	if (strcmp(cip->magic, kLibraryMagic))
@@ -2411,31 +2749,51 @@ FTPGetFiles3(
 	if (pattern1 == NULL)
 		return (kErrBadParameter);
 
+	dstdir2 = NULL;
+	pattern2 = NULL;
+
 	if (dstdir1 == NULL) {
 		dstdir = NULL;
 	} else {
-		dstdir = STRNCPY(dstdir2, dstdir1);
+		dstdir2 = StrDup(dstdir1);
+		if (dstdir2 == NULL) {
+			errRc = kErrMallocFailed;
+			goto return_err;
+		}
 		StrRemoveTrailingLocalPathDelim(dstdir2);
+		dstdir = dstdir2;
+	}
+
+	pattern2 = StrDup(pattern1);
+	if (pattern2 == NULL) {
+		errRc = kErrMallocFailed;
+		goto return_err;
 	}
-	pattern = STRNCPY(pattern2, pattern1);
 	StrRemoveTrailingSlashes(pattern2);
+	pattern = pattern2;
 
 	if (pattern[0] == '\0') {
-		if (recurse == kRecursiveNo)
-			return (kErrBadParameter);
+		if (recurse == kRecursiveNo) {
+			errRc = kErrBadParameter;
+			goto return_err;
+		}
 		pattern = ".";
 		doGlob = kGlobNo;
 	} else if (strcmp(pattern, ".") == 0) {
-		if (recurse == kRecursiveNo)
-			return (kErrBadParameter);
+		if (recurse == kRecursiveNo) {
+			errRc = kErrBadParameter;
+			goto return_err;
+		}
 		doGlob = kGlobNo;
 	}
 	if (recurse == kRecursiveYes)
 		appendflag = kAppendNo;
 
 	batchResult = FTPRemoteGlob(cip, &globList, pattern, doGlob);
-	if (batchResult != kNoErr)
-		return (batchResult);
+	if (batchResult != kNoErr) {
+		errRc = batchResult;
+		goto return_err;
+	}
 
 	cip->cancelXfer = 0;	/* should already be zero */
 
@@ -2564,7 +2922,14 @@ FTPGetFiles3(
 	DisposeLineListContents(&globList);
 	if (batchResult < 0)
 		cip->errNo = batchResult;
-	return (batchResult);
+	errRc = batchResult;
+
+return_err:
+	if (dstdir2 != NULL)
+		free(dstdir2);
+	if (pattern2 != NULL)
+		free(pattern2);
+	return (errRc);
 }	/* FTPGetFiles3 */
 
 
